#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/stat.h>
#include <fcntl.h>

#define FILENAMELEN 1024
#define BUFSZ 1400

int keep_alive = 1;

void stay_dead( int sig ) {
  keep_alive = 0;
  signal( sig, stay_dead );
}

int main( int argc, char* argv[] ) {

  int c;
  struct sockaddr_in addr;
  int sock = 0;
  int timeout = 10;
  int retries = 0;
  int attempts = 0;
  char pidfile[FILENAMELEN];
  char buffer[BUFSZ];
  int sz = 0;
  int waiting = 0;
  int debug = 0;
  FILE* file;
  
  /* default address of warpstar */
  memset( &addr, 0, sizeof(struct sockaddr_in) );
  addr.sin_addr.s_addr = inet_addr( "192.168.0.1" );
  addr.sin_port = htons( 75 );  

  /* default pid file for debian machines */
  strncpy( pidfile, "/var/run/aterm_client.pid", FILENAMELEN ); 
  
  /* get options */
  while( (c = getopt( argc, argv, "s:p:t:r:f:d" )) != -1 ) {
    switch( c ) {

      /* server ip */
    case 's':
      addr.sin_addr.s_addr = inet_addr( optarg );
      break;
      
      /* server port */
    case 'p':
      addr.sin_port = htons( atoi( optarg ) );
      break;

      /* client keep_alive */
    case 't':
      timeout = atoi( optarg );
      break;

      /* connection retries */
    case 'r':
      retries = atoi( optarg );
      break;
      
      /* pid file */
    case 'f':
      strncpy( pidfile, optarg, FILENAMELEN );
      break;

    case 'd':
      debug = 1;
      break;

    default:
    }
  }
  

  /* setup the signal handler */
  signal( SIGINT, stay_dead );
  signal( SIGHUP, stay_dead );
  signal( SIGTERM, stay_dead );

  /* setup the pid file */
  file = fopen( pidfile, "w" );
  if( file == NULL ) {
    perror( "Could not open pid file, aborting." );
    exit( 0 );
  }
  fprintf( file, "%d\n", getpid() );
  fclose( file );

  /* big retry loop */
 RETRY_LOOP:
  while( keep_alive ) {
    
    /* setup the connection */
    addr.sin_family = AF_INET;
    while( ( sock = socket( AF_INET, SOCK_STREAM, 0 ) ) == -1 ) {
      attempts++;
      if( retries == attempts ) {
	perror( "Could not make a socket, giving up." );
	exit( 0 );
      }
      else {
	perror( "Could not make a socket, trying again." );
	sleep( 1 );
      }
    }
    
    /* connect to the server */
    while( connect( sock, (struct sockaddr*) &addr, 
		    sizeof(struct sockaddr_in) ) == -1 ) {
      attempts++;
      if( retries == attempts ) {
	perror( "Could not make a connection, giving up." );
	exit( 0 );
      }
      else {
	perror( "Could not make a connection, trying again." );
	sleep( 1 );
      }
    }
    
    /* send keep_alive timeout */
    sz = snprintf( buffer, BUFSZ, "REG_CLIENT:MODE=TA,TIMEOUT=%d\r\n", timeout );
    if( send( sock, buffer, sz, 0 ) != sz ) {
      attempts++;
      if( retries == attempts ) {
	perror( "Problems sending REG_CLIENT, giving up." );
	close( sock );
	exit( 0 );
      }
      else {
	perror( "Problems sending REG_CLIENT, restarting connection." );
	close( sock );
	sleep( 1 );
	goto RETRY_LOOP;
      }
    }
    
    /* initiate session */
    sz = snprintf( buffer, BUFSZ, "CONNECT_REQ:CHANNEL=01\r\n" );
    if( send( sock, buffer, sz, 0 ) != sz ) {
      attempts++;
      if( retries == attempts ) {
	perror( "Problems sending CONNECT_REQ, giving up." );
	close( sock );
	exit( 0 );
      }
      else {
	perror( "Problems sending CONNECT_REQ, restarting connection." );
	close( sock );
	sleep( 1 );
	goto RETRY_LOOP;
      }
    }
    
    
    while( keep_alive ) {
      
      /* send KEEP_ALIVE */
      sz = snprintf( buffer, BUFSZ, "KEEP_ALIVE\r\n" );
      if( send( sock, buffer, sz, 0 ) != sz ) {
	attempts++;
	if( retries == attempts ) {
	  perror( "Problems sending KEEP_ALIVE, giving up." );
	  close( sock );
	  exit( 0 );
	}
	else {
	  perror( "Problems sending KEEP_ALIVE, restarting connection." );
	  close( sock );
	  sleep( 1 );
	  continue;
	}
      }      

      /* recv, but dont block excessivley*/
      sz = recv( sock, buffer, BUFSZ, MSG_DONTWAIT );

      if( sz == -1 && errno == EAGAIN ) {
	waiting++;
		
	/* if we have waited too long for server then timeout */
	if( waiting == timeout ) {
	  attempts++;
	  if( retries == attempts ) {
	    perror( "Timeout on server KEEP_ALIVE, giving up." );
	    close( sock );
	    exit( 0 );
	  }
	  else {
	    perror( "Timeout on server KEEP_ALIVE, restarting connection." );
	    close( sock );
	    sleep( 1 );
	    goto RETRY_LOOP;
	  }
	}
	
      }
      
      if( sz == -1 && errno != EAGAIN ) {
	attempts++;
	if( retries == attempts ) {
	  perror( "Problems recving KEEP_ALIVEs, giving up." );
	  close( sock );
	  exit( 0 );
	}
	else {
	  perror( "Problems recving KEEP_ALIVEs, restarting connection." );
	  close( sock );
	  sleep( 1 );
	  goto RETRY_LOOP;
	}
      }
	
      /* reset the attempts count, because we actually got something */
      if( sz > 0 ) {
	
	attempts = 0;
	waiting = 0;
	
	if( debug ) {
	  buffer[sz] = '\0';
	  printf( "DEBUG: %s", buffer );
	}

      }
	
      /* wait a sec for the keep_alive */
      sleep( 1 );
      
      
    }
    
    /* skip to the end */
    break;
    
  }

  /* shut down the session */
  sz = snprintf( buffer, BUFSZ, "DISCONNECT_REQ:CHANNEL=01\r\n" );
  send( sock, buffer, sz, 0 );  
  
  /* shut down the connection */
  close( sock );

  /* delete the pid file */
  unlink( pidfile );
  
  return( 0 );

}

